from cycept import jit

import cython
import numpy as np


def test_pass():
    @jit
    def f():
        pass
    assert f() is None

def test_object():
    @jit
    def f(a):
        return a
    a = object()
    assert f(a) is a

def test_type_arg():
    @jit
    def f(a):
        pass
    tps = [int, dict]
    for tp in tps:
        f(tp())
    for tp, key in zip(tps, f.__cycept__.keys()):
        assert key[0] is tp

def test_type_local():
    @jit
    def f(a):
        b = a
    tps = [float, str]
    for tp in tps:
        f(tp())
    for tp, call in zip(tps, f.__cycept__.values()):
        assert call.locals_types['b'] is tp

def test_type_return():
    @jit
    def f(a):
        return a
    tps = [bool, set]
    for tp in tps:
        f(tp())
    for tp, call in zip(tps, f.__cycept__.values()):
        assert call.locals_types['return'] is tp

def test_type_cython():
    @jit
    def f(a):
        b = [a]
        return 2*b.pop()
    f(True)
    types = next(iter(f.__cycept__.values())).types
    assert types == {'a': 'cython.bint', 'b': 'list', 'return': 'cython.Py_ssize_t'}

def test_type_array():
    @jit
    def f(a):
        b = a**0.5
        return b + 1j
    f(np.arange(3))
    call = next(iter(f.__cycept__.values()))
    assert call.locals_types['a'].dtype is np.int64
    assert call.locals_types['b'].dtype is np.float64
    assert call.locals_types['return'].dtype is np.complex128
    assert call.types == {'a': 'cython.longlong[::1]', 'b': 'cython.double[::1]', 'return': 'cython.doublecomplex[::1]'}

def test_type_annotation():
    @jit
    def f(a: int) -> float:
        b : object = [a]
        return 2*b.pop()
    f(True)
    types = next(iter(f.__cycept__.values())).types
    assert types == {'a': 'cython.Py_ssize_t', 'b': 'object', 'return': 'cython.double'}

def test_manual():
    def f(a):
        return 2*a
    g = jit(f)
    assert f(1) == g(1)
    assert next(iter(g.__cycept__.values())).types['return'] == 'cython.Py_ssize_t'

def test_lambda():
    def test(f):
        g = jit(f)
        assert f(1) == g(1)
        assert next(iter(g.__cycept__.values())).types['return'] == 'cython.Py_ssize_t'
    f = lambda a: 2*a
    test(f)
    test(lambda a: 2*a)

def test_closure():
    @jit
    def f(a):
        def g(a):
            return 2*a
        return 1 + g(a)
    for _ in range(2):
        assert f(3) == 7

def test_arg_default():
    @jit
    def f(a, b=3):
        return a + 2*b
    f(1)
    assert f(1) == 7
    assert f(1, 3) == 7
    assert f(1, 5) == 11
    assert f(1, b=5) == 11
    assert f(a=1, b=5) == 11
    assert f(b=5, a=1) == 11

def test_arg_posonly():
    @jit
    def f(a, /, b):
        return a + 2*b
    f(1, 3)
    assert f(1, 3) == 7
    assert f(1, b=3) == 7
    @jit
    def g(a, /, b=3):
        return a + 2*b
    g(1)
    assert g(1) == 7
    assert g(1, 5) == 11
    assert g(1, b=5) == 11

def test_arg_kwonly():
    @jit
    def f(a, *, b):
        return a + 2*b
    f(1, b=3)
    assert f(1, b=3) == 7
    assert f(a=1, b=3) == 7
    assert f(b=3, a=1) == 7
    @jit
    def g(a, *, b=3):
        return a + 2*b
    g(1)
    assert g(1) == 7
    assert g(1, b=5) == 11
    assert g(b=5, a=1) == 11

def test_arg_starargs():
    @jit
    def f(a, *args):
        return a + sum(args)
    f(1)
    assert next(iter(f.__cycept__.values())).locals_types['args'] is tuple
    assert f(1) == 1
    assert f(a=0) == 0
    assert f(1, 2, 3, 4) == 10
    @jit
    def g(*args, a=1):
        return a + sum(args)
    g()
    assert next(iter(g.__cycept__.values())).locals_types['args'] is tuple
    assert g() == 1
    assert g(2, 3, 4) == 10
    assert g(2, 3, 4, a=0) == 9

def test_arg_starstarkwargs():
    @jit
    def f(a, **kwargs):
        return a + kwargs.get('b', 0) + kwargs.get('c', 0)
    f(1)
    assert next(iter(f.__cycept__.values())).locals_types['kwargs'] is dict
    assert f(1) == 1
    assert f(a=0) == 0
    assert f(1, b=2, c=3) == 6
    @jit
    def g(a=1, **kwargs):
        return a + kwargs.get('b', 0) + kwargs.get('c', 0)
    g()
    assert next(iter(g.__cycept__.values())).locals_types['kwargs'] is dict
    assert g() == 1
    assert g(b=2, c=3) == 6
    assert g(b=2, c=3, a=0) == 5

def test_option_none():
    @jit()
    def f():
        pass
    assert f() is None

def test_option_compile():
    for compile in (False, True):
        @jit(compile=compile)
        def f():
            return cython.compiled
        f()  # to compile
        assert f() is compile

def test_option_silent(capfd):
    for silent in (False, True):
        @jit(silent=silent)
        def f():
            pass
        f()
        out, err = capfd.readouterr()
        if silent:
            assert out == ''
        else:
            assert 'Jitting f()' in out  # Cycept
            assert 'Compiling' in out    # Cython
            assert '-O3' in out          # C compiler

def test_option_html():
    for html in (False, True):
        @jit(html=html)
        def f():
            a = 1234
            pass
        f()
        html_annotation = f.__cycept__[()].compiled.html
        if html:
            assert '1234' in html_annotation
        else:
            assert html_annotation is None

def test_option_checks():
    for checks in (False, True):
        @jit(checks=checks)
        def f(a, i):
            return a[i]
        f(np.arange(3), 2)
        source = next(iter(f.__cycept__.values())).source
        deactivated_boundscheck = '@cython.boundscheck(False)' in source
        deactivated_initializedcheck = '@cython.initializedcheck(False)' in source
        assert deactivated_boundscheck is not checks
        assert deactivated_initializedcheck is not checks

def test_option_clike():
    for clike in (False, True):
        @jit(clike=clike)
        def f(a, b):
            return a//b
        for _ in range(2):  # twice in order to use compiled function
            result = f(-1, 2)
        source = next(iter(f.__cycept__.values())).source
        deactivated_wraparoundcheck = '@cython.wraparound(False)' in source
        deactivated_pydivisioncheck = '@cython.cdivision(True)' in source
        assert deactivated_wraparoundcheck is clike
        assert deactivated_pydivisioncheck is clike
        assert result == (0 if clike else -1)

def test_option_array_args():
    def g(a):
        try:
            a *= 2
        except TypeError:
            # Augmented assignment fails for memoryviews
            a[:] = 0
        return a
    for array_args in (False, True):
        @jit(array_args=array_args)
        def f(a):
            return g(a)
        for _ in range(2):  # twice in order to use compiled function
            result = f(np.arange(3))
        assert result.sum() == 6*array_args

def test_option_directives():
    for cdivision in (False, True):
        @jit(directives={'cdivision': cdivision})
        def f(a, b):
            return a//b
        for _ in range(2):  # twice in order to use compiled function
            result = f(-1, 2)
        assert result == (0 if cdivision else -1)

def test_option_optimizations(capfd):
    for level, optimizations in enumerate([False, 1, '-O2', {'-O3': True}]):
        @jit(optimizations=optimizations, silent=False)
        def f():
            pass
        f()
        out, err = capfd.readouterr()
        assert f'-O{level}' in out




import ast
import collections
import contextlib
import distutils
import inspect
import itertools
import os
import pathlib
import re
import sys
import tempfile
import traceback
import typing
import warnings
import webbrowser

import Cython.Build.Cythonize


# Class for storing and dynamically handling
# various aspects of the function call to be jitted.
class FunctionCall:

    class Compiled(typing.NamedTuple):
        func: object
        module: object
        source: str
        html: str

    def __init__(self, func, args, kwargs):
        # Function and call arguments
        self.func = func
        self.args = args
        self.kwargs = kwargs
        # Properties to be lazily constructed
        self._func_name = None
        self._signature = None
        self._arguments = None
        self._arguments_types = None
        self._hash = None
        self._source = None
        self._ast = None
        self._annotations = None
        self._types = None
        self._module = None
        self._module_dict = None
        self._globals = None
        self._nonlocals = None
        # Record of inferred types of local variables
        self.locals_types = {}
        # Additional source lines to be included in the Cython extension module
        self.cython_module_lines = []
        # Compilation products
        self.compiled = None

    # Name of function (with lambda functions renamed)
    @property
    def func_name(self):
        if self._func_name is not None:
            return self._func_name
        self._func_name = self.func.__name__
        if self._func_name == '<lambda>':
            self._func_name = f'_cycept_lambda_{self.hash}'
        return self._func_name

    # Function signature (no type information)
    @property
    def signature(self):
        if self._signature is not None:
            return self._signature
        self._signature = inspect.signature(self.func)
        return self._signature

    # Argument values (dict)
    @property
    def arguments(self):
        if self._arguments is not None:
            return self._arguments
        bound = self.signature.bind(*self.args, **self.kwargs)
        bound.apply_defaults()
        self._arguments = bound.arguments
        return self._arguments

    # Argument types (tuple)
    @property
    def arguments_types(self):
        if self._arguments_types is not None:
            return self._arguments_types
        from .core import get_type
        self._arguments_types = tuple(
            get_type(val) for val in self.arguments.values()
        )
        return self._arguments_types

    # Hash of the function call
    @property
    def hash(self):
        if self._hash is not None:
            return self._hash
        self._hash = abs(hash((self.func, self.arguments_types)))
        return self._hash

    # Source code of the function
    @property
    def source(self):
        if self._source is not None:
            return self._source
        self._source = self.get_source()
        # Separate annotations from source
        self._annotations, self._source = self.extract_annotations()
        # Invalidate AST
        self._ast = None
        return self._source

    @property
    def ast(self):
        if self._ast is not None:
            return self._ast
        self._ast = ast.parse(self.source)
        return self._ast

    # User-defined annotations within the function
    @property
    def annotations(self):
        if self._annotations is not None:
            return self._annotations
        self._annotations, _ = self.extract_annotations()
        return self._annotations

    # Types to be used during compilation
    @property
    def types(self):
        if self._types is not None:
            return self._types
        from .core import convert_to_cython_type
        # Transform recorded types of locals to str
        # representations of corresponding Cython types.
        locals_types = {
            name: convert_to_cython_type(tp)
            for name, tp in self.locals_types.items()
        }
        # Transform annotations to str representations
        # of Cython types.  Keep unrecognized types as is.
        annotations = {
            name: convert_to_cython_type(tp, tp)
            for name, tp in self.annotations.items()
        }
        # Merge recorded types and annotations.
        # On conflict we keep the annotation.
        self._types = locals_types | annotations
        return self._types

    # Module within which function was defined
    @property
    def module(self):
        if self._module is not None:
            return self._module
        self._module = inspect.getmodule(self.func)
        return self._module

    # Module dictionary associated with function module
    @property
    def module_dict(self):
        if self._module_dict is not None:
            return self._module_dict
        self._module_dict = getattr(self.module, '__dict__', None)
        if self._module_dict is None:
            raise ModuleNotFoundError(
                f'Failed to get __dict__ of module {self.module} '
                f'within which function {self.func} is defined'
            )
        return self._module_dict

    # Global names referenced within the function
    @property
    def globals(self):
        if self._globals is not None:
            return self._globals
        self._globals = inspect.getclosurevars(self.func).globals
        return self._globals

    # Non-local names referenced within the function
    @property
    def nonlocals(self):
        if self._nonlocals is not None:
            return self._nonlocals
        self._nonlocals = inspect.getclosurevars(self.func).nonlocals
        return self._nonlocals

    # Method for obtaining the function source
    def get_source(self):
        try:
            source = inspect.getsource(self.func)
        except Exception:
            # inspect.getsource() fails in the interactive REPL.
            # The dill package can hack around this limitation.
            try:
                import dill
            except ModuleNotFoundError:
                warnings.warn(
                    (
                        f'To use cycept.jit interactively, '
                        f'please install the dill Python package:\n'
                        f'    {sys.executable} -m pip install dill\n'
                    ),
                    RuntimeWarning,
                )
                raise
            source = dill.source.getsource(self.func)
        # Dedent source lines
        lines = source.split('\n')
        indentation = ' ' * (len(lines[0]) - len(lines[0].lstrip()))
        if indentation:
            for i, line in enumerate(lines):
                lines[i] = line.removeprefix(indentation)
            source = '\n'.join(lines)
        # Convert lambda function to def function
        if self.func.__name__ == '<lambda>':
            # Extract just the lambda expression
            # (both inspect.getsource() and dill.source.getsource()
            # returns the surrounding code as well in the case of unassigned
            # lambda expressions).
            class LambdaVisitor(ast.NodeVisitor):
                def __init__(self, root):
                    self.root = root
                    self._lambda = None
                def get_lambda(self):
                    self._lambda = None
                    self.visit(self.root)
                    return self._lambda
                def visit_Lambda(self, node):
                    if self._lambda is None:
                        self._lambda = node
            source = ast.unparse(LambdaVisitor(ast.parse(source)).get_lambda())
            # Transform to def function
            indentation = ' ' * 4
            source = re.subn(
                r'.*(^|\W)lambda (.+?): ?',
                rf'def {self.func_name}(\g<2>):\n{indentation}return ',
                source,
                1,
            )[0]
        else:
            # Do AST round-trip, ensuring canonical code style
            # (e.g. 4 space indentation).
            source = ast.unparse(ast.parse(source))
        # Remove jit decorator and decorators above it
        from .core import jit
        lines = source.split('\n')
        for i, line in enumerate(lines):
            if match := re.search(r'^@(.+?)(\(|$)', line):
                deco_name = match.group(1).strip()
                try:
                    deco_func = eval(deco_name, self.module_dict)
                except Exception:
                    continue
                if deco_func is jit:
                    lines = lines[i+1:]
                    break
            elif line.startswith('def '):
                # Arrived at definition line without finding jit decorator.
                # This happens if the "decoration" is done as
                # func = jit(func) rather than using the @jit syntax.
                break
        source = '\n'.join(lines)
        return source

    # Method for extracting user-defined type annotations.
    # The return value is the annotations as well as
    # the modified source without the annotations.
    def extract_annotations(self):
        class AnnotationExtractor(ast.NodeTransformer):
            def __init__(self, call):
                self.call = call
                self.annotations = {}
            def extract(self):
                self.annotations.clear()
                source = ast.unparse(self.visit(self.call.ast))
                return self.annotations, source
            def add_annotation(self, name, tp):
                if not isinstance(name, str):
                    name = ast.unparse(name)
                tp = eval(ast.unparse(tp), self.call.module_dict)
                tp_old = self.annotations.setdefault(name, tp)
                if tp_old == tp:
                    return
                warnings.warn(
                    (
                        f'Name {name!r} annotated with different types '
                        f'in jitted {self.call}: {tp}, {tp_old}. '
                        f'The first will be used'
                    ),
                    RuntimeWarning,
                )
            def visit_arg(self, node):
                if not node.annotation:
                    return node
                self.add_annotation(node.arg, node.annotation)
                return type(node)(**(vars(node) | {'annotation': None}))
            def visit_FunctionDef(self, node):
                node = self.generic_visit(node)
                if not node.returns:
                    return node
                self.add_annotation('return', node.returns)
                return type(node)(**(vars(node) | {'returns': None}))
            def visit_AnnAssign(self, node):
                if node.simple:
                    self.add_annotation(node.target, node.annotation)
                else:
                    warnings.warn(
                        (
                            f'Ignoring complex type annotation '
                            f'{ast.unparse(node)!r} in jitted {self.call}'
                        ),
                        RuntimeWarning,
                    )
                if not node.value:
                    # Annotation-only statement. Remove completely.
                    return None
                # Transform to assignment statement without annotation
                return ast.Assign(**(
                    vars(node) | {'targets': [node.target], 'value': node.value}
                ))
        annotations, source = AnnotationExtractor(self).extract()
        return annotations, source

    # Method for wrapping NumPy array variables in numpy.asarray()
    # whenever operations are used that do not work with Cython memoryviews.
    def protect_arrays(self, array_args):
        from .core import NdarrayTypeInfo
        # Find local NumPy arrays
        names = {
            name
            for name, tp in self.locals_types.items()
            if isinstance(tp, NdarrayTypeInfo)
        }
        names.discard('return')
        if not names:
            # No local arrays found
            return
        # Wrap local arrays
        class ArrayWrapper(ast.NodeTransformer):
            def __init__(self, call, names, array_args, wrapper_func, tmp_name):
                self.call = call
                self.names = names
                self.array_args = array_args
                self.wrapper_func = wrapper_func
                self.tmp_name = tmp_name
                self.shape_attrs = set()
                self.wrapped_any = False
                self.tmp_any = False
            def wrap(self):
                self.shape_attrs.clear()
                self.wrapped_any = False
                self.tmp_any = False
                return self.visit(self.call.ast)
            def wrap_node(self, node):
                if not isinstance(node, ast.Name):
                    return node
                if node.id in self.names:
                    node = ast.Name(**(vars(node) | {'id': f'{self.wrapper_func}({node.id})'}))
                return node
            # Unary and binary operators should always act on NumPy arrays
            def visit_UnaryOp(self, node):
                node = self.generic_visit(node)
                operand = self.wrap_node(node.operand)
                if operand is not node.operand:
                    self.wrapped_any = True
                    node = ast.UnaryOp(**(vars(node) | {'operand': operand}))
                return node
            def visit_BinOp(self, node):
                node = self.generic_visit(node)
                left = self.wrap_node(node.left)
                right = self.wrap_node(node.right)
                if left is not node.left or right is not node.right:
                    self.wrapped_any = True
                    node = ast.BinOp(**(vars(node) | {'left': left, 'right': right}))
                return node
            # Augmented assignment should always be carried out
            # on NumPy arrays.
            def visit_AugAssign(self, node):
                node = self.generic_visit(node)
                target = node.target
                if not isinstance(target, ast.Name) or target.id not in self.names:
                    return node
                self.wrapped_any = True
                self.tmp_any = True
                expr = f'{self.tmp_name} = {self.wrapper_func}({target.id}); {self.tmp_name}'
                target = ast.Name(**(vars(target) | {'id': expr}))
                node.target = target
                return node
            # Use NumPy arrays as arguments within function calls
            # if array_args is True.
            def visit_Call(self, node):
                node = self.generic_visit(node)
                if not self.array_args:
                    return node
                args = [self.wrap_node(arg) for arg in node.args]
                if any(arg_wrapped is not arg for arg, arg_wrapped in zip(args, node.args)):
                    self.wrapped_any = True
                    node = ast.Call(**(vars(node) | {'args': args}))
                return node
            # NumPy arrays and Cython memory views generally share the same
            # attributes. One exception is 'shape', which for memoryviews
            # is meant to be used as memoryview.shape[i]. If the 'shape'
            # attribute is to be used without immediately indexing into it,
            # use a NumPy array instead of a memoryview.
            def visit_Subscript(self, node):
                value = node.value
                if (
                    isinstance(value, ast.Attribute)
                    and isinstance(value.value, ast.Name)
                    and value.value.id in self.names
                    and value.attr == 'shape'
                ):
                    # Record usage of memoryview.shape[i]
                    self.shape_attrs.add(value)
                node = self.generic_visit(node)
                return node
            def visit_Attribute(self, node):
                node = self.generic_visit(node)
                if node not in self.shape_attrs:
                    value = self.wrap_node(node.value)
                    if value is not node.value:
                        self.wrapped_any = True
                        node = ast.Attribute(**(vars(node) | {'value': value}))
                return node
        wrapper_func = '_cycept_asarray_'
        tmp_name = '_cycept_tmp_'
        array_wrapper = ArrayWrapper(self, names, array_args, wrapper_func, tmp_name)
        self._source = ast.unparse(array_wrapper.wrap())
        self._ast = None  # invalidate AST (cannot use the above as expressions are used as names)
        if not array_wrapper.wrapped_any:
            return
        if array_wrapper.tmp_any:
            from .core import record_types
            record_types(self, {tmp_name: None})
        # Add wrapper_func as local variable
        class NodeAdder(ast.NodeTransformer):
            def __init__(self, call, node_to_add):
                self.call = call
                self.node_to_add = node_to_add
            def add_node(self):
                return self.visit(self.call.ast)
            def visit_FunctionDef(self, node):
                if self.node_to_add is None:
                    return node
                node = ast.FunctionDef(**(vars(node) | {'body': [self.node_to_add] + node.body}))
                self.node_to_add = None
                return node
        self._source = ast.unparse(NodeAdder(self, ast.parse(f'{wrapper_func} = _{wrapper_func}_')).add_node())
        self._ast = None  # invalidate AST (above node addition is not strictly legal)
        # We need to assign _{wrapper_func}_ to numpy.asarray() in the extension module
        self.cython_module_lines += [
            '\n# For converting memoryviews back to NumPy arrays',
            f'from numpy import asarray as _{wrapper_func}_'
        ]

    # Method for updating the source from being a Python function
    # to being a Cython extension module.
    def to_cython(self, directives):
        preamble = [
            '#',  # first comment line can have C code attached to it
            f'# Cycept version of {self} with type signature',
            f'# {self!r}',
        ]
        preamble += self.cython_module_lines
        excludes = (self.func_name, 'cython')
        fake_globals = [
            f'{name} = object()'
            for name in collections.ChainMap(self.nonlocals, self.globals)
            if name not in excludes
        ]
        if fake_globals:
            preamble += [
                '\n# Declare fake globals',
                *fake_globals,
            ]
        preamble += [
            '\n# Import Cython',
            'import cython',
            'cimport cython',
        ]
        header = ['\n# Function to be jitted']
        if self.ccall_allowed():
            header.append('@cython.ccall')
        for directive, val in directives.items():
            header.append(f'@cython.{directive}({val!r})')
        excludes = ('return', 'cycept')
        declaration_locals = ', '.join(
            f'{name}={tp}'
            for name, tp in self.types.items()
            if name not in excludes
        )
        declaration_return = self.types.get('return', 'object')
        if declaration_locals:
            header.append(f'@cython.locals({declaration_locals})')
        header.append(f'@cython.returns({declaration_return})')
        self._source = '\n'.join(
            itertools.chain(
                preamble,
                header,
                self.source.split('\n'),
            )
        )

    # Method for determining whether the function to be Cythonized
    # may be done so using @cython.ccall (cpdef).
    def ccall_allowed(self):
        # *args and **kwargs not allowed with ccall
        for param in self.signature.parameters.values():
            if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
                return False
        # Closures not allowed with ccall
        class ClosureVisitor(ast.NodeVisitor):
            def __init__(self, call):
                self.call = call
                self._contains_closure = False
            def contains_closure(self):
                self._contains_closure = False
                self.visit(self.call.ast)
                return self._contains_closure
            def visit_FunctionDef(self, node):
                self.generic_visit(node)
                if node.name != self.call.func_name:
                    self._contains_closure = True
            def visit_Lambda(self, node):
                self._contains_closure = True
        if ClosureVisitor(self).contains_closure():
            return False
        # No violations found
        return True

    # Method for handling Cythonization and C compilation
    def compile(self, optimizations, html, silent):
        # Define context managers for temporarily hack into various
        # objects during Cythonization and compilation.
        @contextlib.contextmanager
        def hack_os_environ():
            cflags_in_environ = ('CFLAGS' in os.environ)
            cflags = os.environ.get('CFLAGS', '')
            os.environ['CFLAGS'] = ' '.join(optimizations)
            yield
            if cflags_in_environ:
                os.environ['CFLAGS'] = cflags
            else:
                os.environ.pop('CFLAGS')
        @contextlib.contextmanager
        def hack_sys_argv():
            module_path = get_module_path(dir_name)
            sys_argv = sys.argv
            sys.argv = list(
                itertools.chain(
                    [''],
                    ['-i', '-3'],
                    ['-q'] * silent,
                    ['-a'] * html,
                    [str(module_path.with_suffix('.pyx'))],
                )
            )
            yield
            sys.argv = sys_argv
        @contextlib.contextmanager
        def hack_sys_path():
            sys.path.append(dir_name)
            yield
            sys.path.remove(dir_name)
        @contextlib.contextmanager
        def hack_distutils_spawn_log(silent):
            class Printer:
                def __init__(self, silent):
                    self.silent = silent
                def print(self, msg, *args, **kwargs):
                    if not self.silent:
                        print(msg)
                def __getattr__(self, attr):
                    return self.print
            printer = Printer(silent)
            module_names = ['spawn', 'dist']
            loggers = {}
            for module_name in module_names:
                module = getattr(distutils, module_name)
                if module_name is None:
                    continue
                logger = getattr(module, 'log')
                if logger is None:
                    continue
                loggers[module_name] = logger
                module.log = printer
            yield
            for module_name, logger in loggers.items():
                module = getattr(distutils, module_name)
                module.log = logger
        # Cythonize and compile extension module within temporary directory,
        # with arguments to Cython and the C compiler provided through hacking
        # of sys.argv and os.environ. The compiled module is then imported
        # (through hacking of sys.path) before it is removed from disk.
        # If not compiling silently, we further hack distutils_spawn_log to
        # print output to stdout.
        module_name = f'_cycept_module_{self.hash}'
        get_module_path = lambda dir_name: pathlib.Path(dir_name) / module_name
        namespace = {}
        html_annotation = None
        with (
            tempfile.TemporaryDirectory() as dir_name,
            hack_os_environ(),
            hack_sys_argv(),
            hack_sys_path(),
            hack_distutils_spawn_log(silent),
        ):
            module_path = get_module_path(dir_name)
            # Write Cython source to file
            module_path.with_suffix('.pyx').write_text(self.source, 'utf-8')
            # Call Cython.Build.Cythonize.main(), which is equivalent
            # to calling the cythonize script.
            ok = True
            try:
                Cython.Build.Cythonize.main()
            except BaseException:
                ok = False
                traceback.print_exc()
            if not ok:
                if sys.flags.interactive:
                    # Do not remove the compilation files immediately when
                    # running interactively, allowing the user to inspect them.
                    input(f'Press Enter to clean up temporary build directory {dir_name} ')
                raise OSError('Cythonization failed')
            # Import function from compiled module into temporary namespace
            exec(f'import {module_name}', namespace)
            # Read in C source and annotated HTML
            source_c = module_path.with_suffix('.c').read_text('utf-8')
            path_html = module_path.with_suffix('.html')
            if path_html.is_file():
                html_annotation = path_html.read_text('utf-8')
        # Extract compiled function
        module_compiled = namespace[module_name]
        module_compiled_dict = module_compiled.__dict__
        func_compiled = module_compiled_dict[self.func_name]
        # Replace fake globals and non-locals with actual
        # globals and non-locals within extension module.
        for name, val in collections.ChainMap(self.nonlocals, self.globals).items():
            if name == self.func_name:
                continue
            module_compiled_dict[name] = val
        # Store compilation products
        self.compiled = self.Compiled(
            func_compiled,
            module_compiled,
            source_c,
            html_annotation,
        )

    # Method for viewing the annotated HTML
    def __call__(self, dir_name=None):
        if self.compiled is None or self.compiled.html is None:
            print(f'No Cython HTML annotation generated for {self.func_name}()')
            return
        if dir_name is None:
            dir_name = tempfile.mkdtemp()  # not cleaned up
        module_path = pathlib.Path(dir_name) / self.compiled.module.__name__
        module_path.with_suffix('.c').write_text(self.compiled.source, 'utf-8')
        path_html = module_path.with_suffix('.html')
        path_html.write_text(self.compiled.html, 'utf-8')
        webbrowser.open_new_tab(str(path_html.as_uri()))
        return dir_name

    # Method for pretty printing (showing types)
    def __repr__(self):
        from .core import prettify_type
        return f'{self.func_name}({{}})'.format(
            ', '.join(
                f'{name}: {prettify_type(tp)}'
                for name, tp in zip(self.arguments, self.arguments_types)
            )
        )

    # Method for pretty printing (showing module)
    def __str__(self):
        return f'function {self.func_name}() defined in {self.module}'




import ast
import collections
import contextlib
import dataclasses
import functools
import pathlib
import re
import sys
import warnings

import cython
import numpy as np

from .call import FunctionCall


# Read current version from pyproject.toml
def get_version():
    version = '?.?.?'
    # Get version from meta data on installed package
    import importlib
    try:
        return importlib.metadata.version('cycept')
    except importlib.metadata.PackageNotFoundError:
        pass
    # Look for pyproject.toml
    def get_tomllib():
        tomllib = None
        try:
            import tomllib
        except ModuleNotFoundError:
            try:
                import toml as tomllib
            except ModuleNotFoundError:
                warnings.warn(
                    (
                        f'For using Cycept with Python < 3.11 please install '
                        f'the toml Python package:\n'
                        f'    {sys.executable} -m pip install toml\n'
                    ),
                    RuntimeWarning,
                )
        return tomllib
    path = pathlib.Path(__file__).resolve().parent
    while True:
        path_pyproject = path / 'pyproject.toml'
        if path_pyproject.is_file():
            tomllib = get_tomllib()
            if tomllib is not None:
                info = tomllib.loads(path_pyproject.read_text('utf-8'))
                version = info['project']['version']
            break
        if path.parent == path:
            break
        path = path.parent
    return version
__version__ = get_version()


# Main JIT decorator function
def jit(func=None, **options):
    """This decorator compiles a Python function to machine code using Cython

    Basic usage
    -----------
        @jit
        def func(x):
            return 2*x

        func(5)

    Arguments
    ---------
        options:
            compile: bool
                Set to False to disable just-in-time compilation.
                Default is True.


            silent: bool
                Set to False to display compilation commands.
                Default value is True.


            html: bool
                Set to True to produce HTML annotations of the compiled code.
                View the HTMl using func.__cycept__().
                Default value is False.


            checks: bool
                Set to True to enable the following checks within Cython:
                    * boundscheck
                    * initializedcheck
                Note that using

                    @jit(checks=True)

                is equivalent to

                    @jit(directives={'boundscheck': True, 'initializedcheck', True})

                Default value is False.


            clike: bool
                Set to True to substitute Python behaviour for
                non-equivalent C behaviour for the following operations:
                    * Indexing with negative integers.
                    * Integer division involving negative numbers.
                    * Exponentiation involving negative numbers.
                If your code does not rely on the above Python specifics,
                explicitly adopting the C behaviour improves performance
                for these operations.

                Note that using

                    @jit(clike=True)

                is equivalent to

                    @jit(directives={'wraparound': False, 'cdivision', True, 'cpow': True})

                Default value is False.


            array_args: bool
                Local NumPy arrays used within the function are always
                converted to Cython memoryviews, allowing for fast indexing.
                If such variables are used as arguments to other functions,
                passing them as memoryviews instead of NumPy arrays may not
                work. To be safe, all memoryview variables are thus converted
                back to NumPy arrays before used as arguments. Set this option
                to False to prevent the conversion back to NumPy arrays.
                Default value is True.


            directives: None | dict
                Allows for adding Cython directives to the transpiled
                function, typically in order to achieve further speedup.
                To e.g. disable boundschecking of indexing operations:

                    @jit(directives={'boundscheck': False})

                You can read about the different directives here:
                https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#compiler-directives


            optimizations: None | str | list |  dict | int | bool
                Specifies which optimizations to use with the C compiler
                (given as CFLAGS).
                If not provided or None, default (aggressive) optimizations
                are used. One or more optimizations can be given as a str
                or list of strs, overwriting the defaults, e.g.

                    @jit(optimizations='-O2')

                If not supplied, default optimizations include
                    * -O3
                    * -ffast-math
                    * -march=native
                To remove e.g. -ffast-math while keeping the remaining
                default optimizations, use

                    @jit(optimizations={'-ffast-math': False})

                If the value within the dict is True rather than False,
                the given optimization is added on top of the default
                optimizations rather than being removed.

                Specifying the optimizations argument as an int 'n'
                is equivalent to passing just the overall optimization level
                (which disables all defaults):

                    @jit(optimizations=n)  # equivalent to the below
                    @jit(optimizations=f'-O{n}')

                If the optimizations argument is given as a bool,
                all optimizations are enabled/disabled. That is,

                    @jit(optimizations=True)

                is equivalent to not passing the optimizations argument
                at all, making use of the defaults. Using

                    @jit(optimizations=False)

                replaces all of the default optimizations with -O0.


    Annotations
    -----------
        Types of variables are automatically inferred at runtime.
        Type annotations can be specified to manually set the types
        of specific variables, e.g.

            @jit
            def func(a: int, b: list) -> float:
                c: object
                c = b[a]
                d: float = (c**1000 % 42)**0.5
                return d

            func(2, list(range(0, 10, 3)))
    """

    if func is None:
        return functools.partial(jit, **options)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Get compiled function or result
        func_compiled, result = transpile(func, wrapper, args, kwargs, **options)
        if func_compiled is not None:
            # Got compiled function. Call it to get result.
            result = func_compiled(*args, **kwargs)
        # If a Cython memoryview is returned, convert to NumPy array
        if 'memoryview' in str(type(result)).lower():
            result = np.asarray(result)
        return result

    # Add storage for function call objects on wrapper
    wrapper.__cycept__ = CyceptStorage()
    return wrapper


# Dict-like class for storing function call objects on the wrapper
class CyceptStorage(dict):
    # When called, dispatch to all function call objects
    def __call__(self):
        if not self:
            print('No compilation has taken place')
        else:
            dir_name = None
            for call in self.values():
                dir_name = call(dir_name)
                if dir_name is None:
                    break


# Function for carrying out the transpilation
# Python → Cython → C → machine code.
def transpile(
    func,
    wrapper,
    args,
    kwargs,
    *,
    compile=True,
    silent=True,
    html=False,
    checks=False,
    clike=False,
    array_args=True,
    directives=None,
    optimizations=None,
):
    if not compile:
        return func, None
    # Fetch function call object, implementing dynamic evaluation
    # of various attributes.
    call = fetch_function_call(func, args, kwargs)
    # If the call has already been transpiled and cached,
    # return immediately.
    if call.compiled is not None:
        return call.compiled.func, None
    # The function call object was not found in cache
    if not silent:
        print(f'Jitting {call!r}')
    # Record types of passed arguments and locals. As a side effect,
    # the function is called and the return value obtained.
    record_types(call)
    result = record_locals(call)
    # Make sure that NumPy arrays are treated as such when necessary
    call.protect_arrays(array_args)
    # Convert Python function source into Cython module source
    directives = get_directives(directives, checks, clike)
    call.to_cython(directives)
    # Cythonize and compile
    optimizations = get_optimizations(optimizations)
    call.compile(optimizations, html, silent)
    # Store the function call object on the wrapper function
    wrapper.__cycept__[call.arguments_types] = call
    # Return the result only
    return None, result


# Function used to fetch FunctionCall instance from the global
# cache or instantiating a new one if not found in the cache.
def fetch_function_call(func, args=None, kwargs=None):
    if isinstance(func, int):
        # Called with function call hash
        return cache[func]
    # We need to instantiate a fresh instance to obtain the hash
    call = FunctionCall(func, args, kwargs)
    call_cached = cache.get(call.hash)
    if call_cached is not None:
        # Found in cache. Disregard the freshly created instance.
        return call_cached
    # Not found in cache. Cache and return the freshly created instance.
    cache[call.hash] = call
    return call
cache = {}


# Function for setting up Cython directives
def get_directives(directives, checks, clike):
    def set_directive(directives, directive, val):
        if not hasattr(cython, directive):
            return
        directives.setdefault(directive, val)
    if directives is None:
        directives = {}
    if not checks:
        checks_disable = {
            'boundscheck': False,
            'initializedcheck': False,
        }
        for directive, val in checks_disable.items():
            set_directive(directives, directive, val)
    if clike:
        clike_enable = {
            'wraparound': False,
            'cdivision': True,
            'cpow': True,
        }
        for directive, val in clike_enable.items():
            set_directive(directives, directive, val)
    for directive, val in directives.items():
        if not hasattr(cython, directive):
            warnings.warn(
                f'Cython directive {directive!r} does not seem to exist',
                RuntimeWarning,
            )
    return directives


# Function for setting up compiler optimizations (CFLAGS)
def get_optimizations(optimizations):
    get_defaults = lambda: optimizations_default.copy()
    if optimizations is None:
        # No optimizations provided. Use defaults.
        optimizations = get_defaults()
    elif isinstance(optimizations, (bool, int)):
        # Enable or disable all optimizations
        if optimizations is True:
            optimizations = get_defaults()
        else:
            level = abs(int(optimizations))
            optimizations = [f'-O{level}']
    elif isinstance(optimizations, str):
        # Transform str of optimizations to list
        optimizations = optimizations.split(' ')
    elif isinstance(optimizations, dict):
        # Use defaults with amendments
        optimizations_ammended = get_defaults()
        for optimization, enable in optimizations.items():
            if enable and optimization not in optimizations_ammended:
                optimizations_ammended.append(optimization)
            elif not enable and optimization in optimizations_ammended:
                optimizations_ammended.remove(optimization)
        optimizations = optimizations_ammended
    else:
        # Use supplied optimizations as is
        optimizations = list(optimizations)
    return optimizations
optimizations_default = [
    '-DNDEBUG',
    '-O3',
    '-funroll-loops',
    '-ffast-math',
    '-march=native',
]


# Function for extracting the Python/NumPy type off of a value
def get_type(val):
    tp = construct_ndarray_type_info(val)
    if tp is not None:
        return tp
    return type(val)


# Function responsible for constructing NdarrayTypeInfo instances,
# storing type information of arrays/memoryviews.
def construct_ndarray_type_info(a):
    def construct(dtype, ndim, c_contig, f_contig):
        if c_contig:
            # Disregard F-contiguousness if C-contiguous
            f_contig = False
        return NdarrayTypeInfo(dtype, ndim, c_contig, f_contig)
    if isinstance(a, NdarrayTypeInfo):
        return a
    if isinstance(a, numpy_array_type):
        return construct(
            a.dtype.type,
            a.ndim,
            a.flags['C_CONTIGUOUS'],
            a.flags['F_CONTIGUOUS'],
        )
    if isinstance(a, cython_array_type):
        return construct(
            cython_types_reverse[_cython_types[a.dtype]],
            a.ndim,
            a.is_c_contig,
            a.is_f_contig,
        )
    # Passed value is not an array
    return None
numpy_array_type = np.ndarray
cython_array_type = type(cython.int[:])


# Type used to discern different kinds of arrays (memoryviews)
@dataclasses.dataclass(frozen=True)
class NdarrayTypeInfo:
    dtype: np.dtype.type
    ndim: int
    c_contig: bool
    f_contig: bool

    @property
    def dtype_cython(self):
        return convert_to_cython_type(self.dtype)

    @property
    def __name__(self):
        name = '{}[{}]'.format(self.dtype_cython, ', '.join(':' * self.ndim))
        if self.c_contig:
            name = name.replace(':]', '::1]')
        elif self.f_contig:
            name = name.replace('[:', '[::1')
        return name

    def __str__(self):
        return self.__name__.removeprefix('cython.')


# Function for recording types
def record_types(call, variables=None):
    if isinstance(call, int):
        # Called with function call hash
        call = fetch_function_call(call)
    if variables is None:
        # Use function call arguments
        variables = call.arguments
    for name, val in variables.items():
        tp = get_type(val)
        tp_old = call.locals_types.setdefault(name, tp)
        if tp_old == tp:
            continue
        if (
            (tp_old in (float, int, bool) and tp in (int, bool))
            or (tp_old == complex and tp in (float, int, bool))
        ):
            # Allow implicit upcasting of Python scalars
            # as well as casting back and forth between bools and ints.
            call.locals_types.record[name] = tp_old
        else:
            warnings.warn(
                (
                    f'Name {name!r} assigned different types '
                    f'in jitted {call}: {tp}, {tp_old}. '
                    f'The first will be used'
                ),
                RuntimeWarning,
            )


# Function for obtaining the types of locals within a function
def record_locals(call):
    # Create copy of function source with the function renamed
    class FunctionRenamer(ast.NodeTransformer):
        def __init__(self, name_old, name_new):
            self.name_old = name_old
            self.name_new = name_new
            self.visit_Name = self.make_visitor('id')
            self.visit_FunctionDef = self.make_visitor('name')
        def make_visitor(self, attr):
            def rename(node):
                node = self.generic_visit(node)
                if getattr(node, attr) == self.name_old:
                    node = type(node)(**(vars(node) | {attr: self.name_new}))
                return node
            return rename
    func_name_tmp = f'_cycept_func_{call.hash}'
    source = ast.unparse(
        FunctionRenamer(call.func_name, func_name_tmp).visit(call.ast)
    )
    call._ast = None  # invalidate AST after renaming
    # Add type recording calls to source
    cycept_module_refname = '__cycept__'
    record_str = f'{cycept_module_refname}.record_types({call.hash}, locals())'
    source = re.sub(
        r'( |;)return($|\W)',
        rf'\g<1>{record_str}; return\g<2>',
        source,
    )
    indentation = ' ' * 4
    source += f'\n{indentation}{record_str}'
    # Define and call modified function within definition module,
    # with recording of the types added as a side effect.
    # Temporarily add a reference to this module on the module of
    # the function to be jitted. Also add non-local objects referenced by
    # the function directly to the module.
    @contextlib.contextmanager
    def hack_module_dict():
        call.module_dict[cycept_module_refname] = sys.modules['cycept.core']
        nonlocals_ori = {}
        for name, val in call.nonlocals.items():
            if name in call.module_dict:
                nonlocals_ori[name] = call.module_dict[name]
            call.module_dict[name] = val
        yield
        call.module_dict.pop(cycept_module_refname)
        for name in call.nonlocals:
            call.module_dict.pop(name)
        for name, val in nonlocals_ori.items():
            call.module_dict[name] = val
    with hack_module_dict():
        # Define modified function within definition module
        exec(source, call.module_dict)
        # Call modified function with passed arguments
        return_val = call.module_dict[func_name_tmp](*call.args, **call.kwargs)
    # Remove modified function from definition module
    call.module_dict.pop(func_name_tmp)
    # Add return type to record
    record_types(call, {'return': return_val})
    # Return the return value from the function call
    return return_val


# Function returning pretty str representation of type
def prettify_type(tp):
    if isinstance(tp, NdarrayTypeInfo):
        return str(tp)
    tp_name = getattr(tp, '__name__', getattr(tp, 'name', None))
    if tp_name is not None:
        return tp_name
    return str(tp)


# Function for converting a Python/NumPy/Cython type to the
# str representation of the corresponding Cython type.
def convert_to_cython_type(tp, default=object):
    def convert_array_type(tp):
        if cython_types[tp.dtype] == 'object':
            return 'object'
        return tp.__name__
    # Type already given as str
    if isinstance(tp, str):
        return repr(tp)
    # NumPy array or Cython memoryview
    tp_arr = construct_ndarray_type_info(tp)
    if tp_arr is not None:
        return convert_array_type(tp_arr)
    # Python/NumPy/Cython
    tp_str = cython_types.get(tp)
    if tp_str is not None:
        return tp_str
    # Unrecognized
    if default is not None and tp is not default:
        return convert_to_cython_type(default, None)
    tp_str = getattr(tp, '__name__', getattr(tp, 'name', None))
    if tp_str is not None:
        return tp_str
    return str(tp)


# Mapping of Python/NumPy types to str representations
# of corresponding Cython types.
cython_default_integral = 'cython.Py_ssize_t'  # typically 64-bit
cython_types = collections.defaultdict(
    lambda: 'object',
    {
        # Python scalars
        bool: 'cython.bint',
        int: cython_default_integral,
        float: 'cython.double',
        complex: 'cython.complex',
        # NumPy signed integral scalars
        np.int8: 'cython.schar',
        np.int16: 'cython.short',
        np.int32: 'cython.int',
        np.int64: 'cython.longlong',
        # NumPy unsigned integral scalars
        np.uint8: 'cython.uchar',
        np.uint16: 'cython.ushort',
        np.uint32: 'cython.uint',
        np.uint64: 'cython.ulonglong',
        # NumPy floating scalars
        np.float32: 'cython.float',
        np.float64: 'cython.double',
        np.longdouble: 'cython.longdouble',  # platform dependent
        # NumPy complex floating scalars
        np.complex64: 'cython.floatcomplex',  # Cython type may not be available at compile time
        np.complex128: 'cython.doublecomplex',
        np.complex256: 'cython.longdoublecomplex',  # Cython type may not be available at compile time; np.longdoublecomplex not defined
        # Python containers
        **{
            tp: tp.__name__
            for tp in [
                bytearray,
                bytes,
                dict,
                frozenset,
                list,
                object,
                set,
                str,
                tuple,
            ]
        },
    },
)
# Also create reverse mapping
cython_types_reverse = collections.defaultdict(
    lambda: object,
    {name: tp for tp, name in cython_types.items()},
)
# Now add the Cython types themselves
cython_types |= {
    getattr(cython, name.removeprefix('cython.')): name
    for tp, name in cython_types.items()
    if name.startswith('cython.')
}

